# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Any, Optional

try:
    import tomllib
except ModuleNotFoundError:  # pragma: no cov
    import pip._vendor.tomli as tomllib

import yaml


def load_pyproject(file: Optional[str] = None) -> dict[str, Any]:
    """Load Configuration from pyproject.toml file.

    :param file: A file path or string path that keeping the pyproject.toml.

    :rtype: dict[str, Any]
    """
    f: str = file or "./pyproject.toml"
    pyproject: Path = Path(f)

    if not pyproject.exists():
        return {}

    with pyproject.open(mode="rb") as f:
        return tomllib.load(f)


def load_config() -> dict[str, Any]:
    """Return config of the shelf package that was set on pyproject.toml.

    :rtype: dict[str, Any]
    """
    data: dict[str, Any] = {}

    conf_file: Path = Path(".clishelf.yaml")
    if conf_file.exists():
        data = yaml.safe_load(conf_file.read_text(encoding="utf-8"))

    return {
        **data,
        **load_pyproject().get("tool", {}).get("shelf", {}),
    }


class Bcolors(str, Enum):
    """An Enum for colors using ANSI escape sequences.

    Reference:
        - https://stackoverflow.com/questions/287871
    """

    HEADER: str = "\033[95m"
    OK_BLUE: str = "\033[94m"
    OK_CYAN: str = "\033[96m"
    OK_GREEN: str = "\033[92m"
    WARNING: str = "\033[93m"
    FAIL: str = "\033[91m"
    END: str = "\033[0m"
    BOLD: str = "\033[1m"
    OK: str = "\033[92m"
    INFO: str = "\033[94m"
    ERROR: str = "\033[91m"


class Level(str, Enum):
    """An Enum for notification levels that relate with Bcolors enum object."""

    OK: str = "OK"
    INFO: str = "INFO"
    WARNING: str = "WARNING"
    ERROR: str = "ERROR"


def make_color(message: str, level: Level) -> str:
    """Print the message with a color for the corresponding level.

    :param message: A message string that want to echo.
    :param level: A level of color.

    :rtype: str
    """
    return (
        f"{Bcolors[level].value}{Bcolors.BOLD.value}{level.value}: "
        f"{message}{Bcolors.END.value}"
    )


@dataclass(frozen=True)
class Profile:
    """Profile dataclass object for keeping a Git profile data."""

    name: str
    email: str



from datetime import datetime
from unittest.mock import patch

import pytest

import clishelf.settings as settings


def test_bump_version_setting():
    bump_setting = settings.BumpVerConf

    assert bump_setting.regex == bump_setting.get_regex()
    assert bump_setting.regex_dt == bump_setting.get_regex(is_dt=True)

    update_dt_pre = bump_setting.update_dt_pre
    assert update_dt_pre("20240102") == "20240102.1"
    assert update_dt_pre("20240102.5") == "20240102.6"

    with pytest.raises(ValueError):
        bump_setting.update_dt_pre("202401.post")


def test_bump_version_get_version():
    bump_setting = settings.BumpVerConf
    rs = bump_setting.get_version(
        version=3,
        params={
            "version": "",
            "changelog": "",
            "file": "",
        },
    )
    assert rs == bump_setting.v1.format(
        changelog="",
        main=bump_setting.main.format(
            version="",
            msg=bump_setting.msg,
            regex=bump_setting.regex,
            file="",
        ),
    )


@patch("clishelf.settings.datetime.datetime")
def test_bump_version_get_version_dt(mock_datetime):
    mock_datetime.now.return_value = datetime(2024, 1, 1, 0, 0)
    bump_setting = settings.BumpVerConf
    rs = bump_setting.get_version(
        version=2,
        params={
            "version": "",
            "changelog": "",
            "file": "",
            "action": "date",
        },
        is_dt=True,
    )
    assert rs == bump_setting.v2.format(
        changelog="",
        main=bump_setting.main_dt.format(
            version="",
            new_version="20240101",
            msg=bump_setting.msg,
            regex=bump_setting.regex_dt,
            file="",
        ),
    )

    rs = bump_setting.get_version(
        version=2,
        params={
            "version": "20240101",
            "changelog": "",
            "file": "",
            "action": "pre",
        },
        is_dt=True,
    )
    assert rs == bump_setting.v2.format(
        changelog="",
        main=bump_setting.main_dt.format(
            version="20240101",
            new_version="20240101.1",
            msg=bump_setting.msg,
            regex=bump_setting.regex_dt,
            file="",
        ),
    )

    with pytest.raises(ValueError):
        bump_setting.get_version(
            version=2,
            params={
                "version": "20240101",
                "changelog": "",
                "file": "",
                "action": "post",
            },
            is_dt=True,
        )



# ------------------------------------------------------------------------------
# Copyright (c) 2022 Korawich Anuttra. All rights reserved.
# Licensed under the MIT License. See LICENSE in the project root for
# license information.
# ------------------------------------------------------------------------------
from collections.abc import Iterator
from datetime import datetime
from pathlib import Path
from textwrap import dedent
from unittest.mock import DEFAULT, patch

from clishelf.git import CommitLog, CommitMsg, Profile
from clishelf.version import (
    create_changelog,
    current_version,
    get_changelog,
    map_group_commit_logs,
    write_bump_file,
    write_group_log,
)


def side_effect_func(*args, **kwargs):
    _ = kwargs
    if ".bumpversion.cfg" in args[0]:
        return Path(__file__).parent / ".bumpversion.cfg"
    elif "__version__.py" in args[0]:
        return Path(__file__).parent / "__version__.py"
    return DEFAULT


@patch("clishelf.git.get_commit_logs")
def test_map_group_commit_logs(mock_get_commit_logs):
    commit_log = CommitLog(
        hash="f477e87",
        refs="HEAD",
        date=datetime(2024, 1, 1),
        msg=CommitMsg(":toolbox: build: add coverage workflow"),
        author=Profile(name="test", email="test@mail.com"),
    )
    mock_get_commit_logs.return_value = iter([commit_log])

    assert map_group_commit_logs() == {
        "HEAD": {"Build & Workflow": [commit_log]}
    }


def test_get_changelog(test_path: Path):
    changelog_file: Path = test_path / "test_changelog.md"
    with changelog_file.open(mode="w") as f:
        f.writelines(
            [
                "# Changelogs\n\n",
                "## Latest Changes\n\n",
                "## 0.0.2\n\n",
                "### Features\n\n",
                "- :dart: feat: second commit (_2024-01-02_)\n\n",
                "## 0.0.1\n\n",
                "### Features\n\n",
                "- :dart: feat: first initial (_2024-01-01_)\n",
            ]
        )
    rs: Iterator[str] = get_changelog(changelog_file)
    assert list(rs) == [
        "# Changelogs",
        "",
        "## Latest Changes",
        "",
        "## 0.0.2",
        "",
        "### Features",
        "",
        "- :dart: feat: second commit (_2024-01-02_)",
        "",
        "## 0.0.1",
        "",
        "### Features",
        "",
        "- :dart: feat: first initial (_2024-01-01_)",
    ]

    rs = get_changelog(changelog_file, refresh=True)
    assert list(rs) == ["# Changelogs", "", "## Latest Changes"]

    rs = get_changelog(changelog_file, tags=["0.0.2"], refresh=True)
    assert list(rs) == ["# Changelogs", "", "## Latest Changes", "", "## 0.0.2"]

    changelog_file.unlink()


@patch("clishelf.version.map_group_commit_logs")
def test_create_changelog(mock_map_group_commit_logs):
    commit_logs: list[CommitLog] = [
        CommitLog(
            hash="f477e87",
            refs="HEAD",
            date=datetime(2024, 1, 2),
            msg=CommitMsg(":toolbox: build: add coverage workflow"),
            author=Profile(name="test", email="test@mail.com"),
        ),
        CommitLog(
            hash="dc25a22",
            refs="HEAD",
            date=datetime(2024, 1, 1),
            msg=CommitMsg(":construction: refactored: add new features"),
            author=Profile(name="test", email="test@mail.com"),
        ),
    ]
    mock_map_group_commit_logs.return_value = {
        "HEAD": {
            "Build & Workflow": [commit_logs[0]],
            "Code Changes": [commit_logs[1]],
        },
    }
    write_changelog_file = Path(__file__).parent / "test_write_changelog.md"
    with write_changelog_file.open(mode="w") as f:
        f.writelines(
            [
                "# Changelogs\n\n",
                "## Latest Changes\n\n",
                "## 0.0.1\n\n",
                "### Features\n\n",
                "- :dart: feat: first initial (_2024-01-01_)\n\n",
                "## NOTED\n\n",
                "This line should comment for EOF\n\n",
            ]
        )

    create_changelog(write_changelog_file, all_tags=True)

    assert write_changelog_file.exists()
    assert write_changelog_file.read_text().replace(" ", "") == dedent(
        """# Changelogs

        ## Latest Changes

        ### :black_nib: Code Changes

        - :construction: refactored: add new features (_2024-01-01_)

        ### :package: Build & Workflow

        - :toolbox: build: add coverage workflow (_2024-01-02_)

        ## 0.0.1

        ### Features

        - :dart: feat: first initial (_2024-01-01_)

        ## NOTED

        This line should comment for EOF

        """.replace(
            " ", ""
        )
    )
    write_changelog_file.unlink()


def test_write_group_log():
    test_file_path: Path = Path(__file__).parent / "test_write_group_log.md"
    group_log = {
        "Build & Workflow": [
            CommitLog(
                hash="f477e87",
                refs="HEAD",
                date=datetime(2024, 1, 2),
                msg=CommitMsg(":toolbox: build: add coverage workflow"),
                author=Profile(name="test", email="test@mail.com"),
            )
        ]
    }
    with test_file_path.open(mode="w", newline="") as f:
        write_group_log(f, group_log, "HEAD")

    assert test_file_path.exists()
    assert test_file_path.read_text().replace(" ", "") == dedent(
        """## HEAD

        ### :package: Build & Workflow

        - :toolbox: build: add coverage workflow (_2024-01-02_)

        """.replace(
            " ", ""
        )
    )

    test_file_path.unlink()

    group_log = {
        "Build & Workflows": [
            CommitLog(
                hash="f477e87",
                refs="HEAD",
                date=datetime(2024, 1, 2),
                msg=CommitMsg(":toolbox: build: add coverage workflow"),
                author=Profile(name="test", email="test@mail.com"),
            )
        ]
    }
    with test_file_path.open(mode="w", newline="") as f:
        write_group_log(f, group_log, "HEAD")

    assert test_file_path.exists()
    assert test_file_path.read_text().replace(" ", "") == dedent(
        """## HEAD\n""".replace(" ", "")
    )

    test_file_path.unlink()


@patch("clishelf.version.Path", side_effect=side_effect_func)
def test_write_bump_file(mock_path):
    bump_file_path: Path = Path(__file__).parent / ".bumpversion.cfg"

    write_bump_file(
        param={
            "version": "",
            "changelog": "",
            "file": "__about__.py",
        },
    )

    assert bump_file_path.exists()

    bump_file_path.unlink()


@patch("clishelf.version.Path", side_effect=side_effect_func)
def test_current_version(mock_path):
    version_file_path: Path = Path(__file__).parent / "__version__.py"

    with version_file_path.open(mode="w") as f:
        f.writelines(["__version__ = 0.0.1\n", "__version_dt__ = 20240101\n"])

    assert current_version("__version__.py") == "0.0.1"
    assert current_version("__version__.py", is_dt=True) == "20240101"

    version_file_path.unlink()



import datetime
import sys
import unittest
from unittest.mock import DEFAULT, patch

import clishelf.git as git
from clishelf.emoji import demojize


def side_effect_func(*args, **kwargs):
    if any(["git", "config", "--local", "user.name"] == a for a in args):
        _ = kwargs
        return "Test User".encode(encoding=sys.stdout.encoding)
    elif any(["git", "config", "--local", "user.email"] == a for a in args):
        _ = kwargs
        return "test@mail.com".encode(encoding=sys.stdout.encoding)
    else:
        return DEFAULT


def side_effect_bn_tg_func(*args, **kwargs):
    if any(["git", "rev-parse", "--abbrev-ref", "HEAD"] == a for a in args):
        _ = kwargs
        return "0.1.2".encode(encoding=sys.stdout.encoding)
    elif any(["git", "describe", "--tags", "--abbrev=0"] == a for a in args):
        _ = kwargs
        return "v0.0.1".encode(encoding=sys.stdout.encoding)
    else:
        return DEFAULT


class GitModelTestCase(unittest.TestCase):
    def test_commit_prefix_model(self):
        rs = git.CommitPrefix(
            name="test",
            group="A",
            emoji=":dart:",
        )
        self.assertEqual(hash(rs), hash(rs.name))
        self.assertEqual("test", str(rs))

    def test_commit_prefix_group_model(self):
        rs = git.CommitPrefixGroup(
            name="test",
            emoji=":dart:",
        )
        self.assertEqual(hash(rs), hash(rs.name))
        self.assertEqual("test", str(rs))

    def test_commit_message_model(self):
        msg = git.CommitMsg(
            content=":dart: feat: start initial testing",
            mtype=None,
        )
        self.assertEqual(
            "Features: :dart: feat: start initial testing", str(msg)
        )

        msg = git.CommitMsg(
            content=":dart: demo: start initial testing",
            mtype=None,
        )
        self.assertEqual(
            "Code Changes: :dart: demo: start initial testing", str(msg)
        )

        msg = git.CommitMsg(
            content=":dart: start initial testing",
            mtype=None,
        )
        self.assertEqual("Code Changes: :dart: start initial testing", str(msg))

        msg = git.CommitMsg(
            content="⬆️ deps: upgrade dependencies from main branch (#63)",
            mtype=None,
        )
        self.assertEqual(
            (
                "Dependencies: :arrow_up: deps:  upgrade dependencies from "
                "main branch (#63)"
            ),
            str(msg),
        )

        msg = git.CommitMsg(
            content="Merge branch 'main' into dev",
            mtype=None,
        )
        self.assertEqual(
            "Code Changes: :fast_forward: merge: branch 'main' into dev",
            str(msg),
        )

    def test_commit_message_model_raise(self):
        with self.assertRaises(ValueError):
            git.CommitMsg(
                content="demo: start initial testing",
                mtype=None,
            )

    @patch("clishelf.utils.load_pyproject")
    def test_commit_message_model_raise_without_conf(self, mock_load_pyproject):
        with self.assertRaises(ValueError):
            mock_load_pyproject.return_value = {
                "tool": {"shelf": {"git": {"commit_prefix_force_fix": False}}},
            }
            git.CommitMsg(
                content="⬆️ demo: start initial testing",
                mtype=None,
            )

    def test_commit_log_model(self):
        log = git.CommitLog(
            hash="test",
            refs="HEAD",
            date=datetime.date(2023, 1, 1),
            msg=git.CommitMsg(":dart: feat: start initial testing"),
            author=git.Profile(
                name="test",
                email="test@mail.com",
            ),
        )
        self.assertEqual(
            (
                "test|2023-01-01|:dart: feat: start initial testing|"
                "test|test@mail.com|HEAD"
            ),
            str(log),
        )


@patch("clishelf.git.subprocess.check_output", side_effect=side_effect_func)
@patch("clishelf.utils.load_pyproject")
def test_load_profile(mock_load_pyproject, mock):
    mock_load_pyproject.return_value = {}
    rs = git.load_profile()

    assert mock.called
    assert isinstance(rs, git.Profile)
    assert "Test User" == rs.name
    assert "test@mail.com" == rs.email


def test_get_commit_prefix():
    assert 28 == len(list(git.get_commit_prefix()))


def test_get_commit_prefix_group():
    data: tuple[git.CommitPrefixGroup, ...] = git.get_commit_prefix_group()
    feat: git.CommitPrefixGroup = [cm for cm in data if cm.name == "Features"][
        0
    ]
    assert ":tada:" == feat.emoji


@patch(
    "clishelf.git.subprocess.check_output",
    side_effect=side_effect_bn_tg_func,
)
def test_get_latest_tag(mock):
    result = git.get_latest_tag()

    assert mock.called
    assert "v0.0.1" == result


def test_git_demojize():
    assert "test :fire: :fire:" == demojize(
        "test 🔥 :fire:", emojis=git.get_git_emojis()
    )


def test_validate_commit_msg_warning():
    rs = git._validate_commit_msg_warning([":dart: feat: demo", ""])
    assert rs == [
        (
            "There should be between 21 and 50 characters in the "
            "commit title."
        ),
        "There should at least 3 lines in your commit message.",
        "There should not has dot in the end of commit message.",
    ]

    rs = git._validate_commit_msg_warning(
        [":dart: feat: demo test validate for warning.", "empty"]
    )
    assert rs == [
        "There should at least 3 lines in your commit message.",
        ("There should be an empty line between the commit title " "and body."),
    ]

    rs = git._validate_commit_msg_warning(
        [
            ":dart: feat: demo test validate for warning.",
            "",
            "body of commit log",
            "",
        ]
    )
    assert rs == []


def test_validate_commit_msg():
    rs = git.validate_commit_msg([])
    assert rs == (
        ["Please supply commit message without start with ``#``."],
        git.Level.ERROR,
    )

    rs = git.validate_commit_msg(
        [
            ":dart: feat: demo test validate for warning.",
            "",
            "body of commit log",
            "",
        ]
    )
    assert rs == (
        ["The commit message has the required pattern."],
        git.Level.OK,
    )

    rs = git.validate_commit_msg(
        [
            ":dart: feat: demo test validate for warning.",
            "",
            (
                "body of commit log that has character more that 72 and "
                "it will return some warning message from function"
            ),
            "",
        ]
    )
    assert rs == (
        ["The commit body should wrap at 72 characters at line: 3."],
        git.Level.WARNING,
    )
